/*
 * Copyright (c) 2018 Samsung Electronics Co., Ltd.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

#include <vector>

#include "idlc/c_gen/c_body_gen_base.h"

namespace {
#include "idlc/c_gen/c_body_gen_base_cb.h"
}

namespace tidl {

CBodyGeneratorBase::CBodyGeneratorBase(std::shared_ptr<Document> doc)
    : CGeneratorBase(doc) {
  parcel_type_map_ = {
    {"char", "byte"},
    {"int", "int32"},
    {"short", "int16"},
    {"long", "int64"},
    {"string", "string"},
    {"bool", "bool"},
    {"float", "float"},
    {"double", "double"},
    {"bundle", "bundle"},
  };
}

void CBodyGeneratorBase::GenStructures(std::ofstream& stream) {
  for (auto& i : GetDocument().GetBlocks()) {
    if (i->GetType() == Block::TYPE_STRUCTURE) {
      const Structure &st = static_cast<const Structure&>(*i);
      GenStructure(stream, st);
      for (auto& j : st.GetElements().GetElms()) {
       auto& t = j->GetType();
       AddStructureFromType(t);
      }
    } else {
      const Interface &inf = static_cast<const Interface&>(*i);
      for (auto& d : inf.GetDeclarations().GetDecls()) {
        for (auto& p : d->GetParameters().GetParams()) {
          if (IsDelegateType(inf, p->GetParameterType().GetBaseType()))
            continue;
          AddStructureFromType(p->GetParameterType().GetBaseType());
        }
      }
    }
  }

  for (auto& p : GetStructures()) {
    const Structure& st = *p.second;
    GenStructure(stream, st);
  }
}

void CBodyGeneratorBase::GenStructure(std::ofstream& stream,
                                      const Structure& st) {
  GenStructureDeclaration(stream, st);
  GenStructureParcelSerializer(stream, st);
  GenStructureParcelDeserializer(stream, st);
  GenStructureConstructor(stream, st);
  GenStructureDestructor(stream, st);
  GenStructureCloner(stream, st);
  GenStructureSetter(stream, st);
  GenStructureGetter(stream, st);
  GenStructureIterator(stream, st);
}

void CBodyGeneratorBase::GenStructureDeclaration(std::ofstream& stream,
                                                 const Structure& st) {
  stream << SmartIndent(GenTemplateString(CB_STRUCT_DECL,
        [&]()->std::string {
          return st.GetComments();
        },
        [&]()->std::string {
          return GetStructIdWithNamespace(st);
        },
        [&]()->std::string {
          std::string str;
          for (auto& i : st.GetElements().GetElms()) {
            str += NLine(1) + GetStringFromElementType(i->GetType()) +
                i->GetID() + ";";
            if (i->GetType().ToString() == "array")
              str += NLine(1) + "int " + i->GetID() + "_size;";
          }
          return str;
        }));
}

void CBodyGeneratorBase::GenStructureParcelSerializer(std::ofstream& stream,
                                              const Structure& st) {
  stream << SmartIndent(GenTemplateString(
        ReplaceAll(CB_STRUCT_SERIALIZER, "##", GetStructIdWithNamespace(st)),
        [&]()->std::string {
          std::string str;
          for (auto& i : st.GetElements().GetElms()) {
            str += NLine(1);
            str += GetParcelWriteString(i->GetID(), i->GetType());
          }
          return str;
        }));
}

void CBodyGeneratorBase::GenStructureParcelDeserializer(std::ofstream& stream,
                                                const Structure& st) {
  stream << SmartIndent(GenTemplateString(
        ReplaceAll(CB_STRUCT_DESERIALIZER, "##", GetStructIdWithNamespace(st)),
        [&]()->std::string {
          std::string str;
          for (auto& i : st.GetElements().GetElms()) {
            str += NLine(1);
            str += GetParcelReadString(i->GetID(), i->GetType());
          }
          return str;
        }));
}

void CBodyGeneratorBase::GenStructureConstructor(std::ofstream& stream,
                                                 const Structure& st) {
  stream << SmartIndent(ReplaceAll(CB_STRUCT_CTOR, "##",
      GetStructIdWithNamespace(st)));
}

void CBodyGeneratorBase::GenStructureDestructor(std::ofstream& stream,
                                                const Structure& st) {
  stream << SmartIndent(GenTemplateString(
        ReplaceAll(CB_STRUCT_DTOR, "##",
            GetStructIdWithNamespace(st)),
        [&]()->std::string {
          std::string str;
          for (auto& i : st.GetElements().GetElms()) {
            str += GetFinalizeString(i->GetID(), i->GetType());
          }
          return str;
        }));
}

void CBodyGeneratorBase::GenStructureSetter(std::ofstream& stream,
                                            const Structure& st) {
  for (auto& i : st.GetElements().GetElms()) {
    stream << SmartIndent(GenTemplateString(CB_STRUCT_SETTER,
          [&]()->std::string {
            return GetStructIdWithNamespace(st);
          },
          [&]()->std::string {
            if (i->GetType().ToString() == "list")
              return "add";
            return "set";
          },
          [&]()->std::string {
            return i->GetID();
          },
          [&]()->std::string {
            return GetStructIdWithNamespace(st);
          },
          [&]()->std::string {
            if (i->GetType().IsUserDefinedType())
              return GetParcelParamTypeString(i->GetType());

            if (i->GetType().ToString() == "list") {
              if (i->GetType().GetMetaType()->IsUserDefinedType() ||
                  i->GetType().GetMetaType()->ToString() == "list" ||
                  i->GetType().GetMetaType()->ToString() == "array") {
                return GetParcelParamTypeString(*i->GetType().GetMetaType());
              } else {
                return ConvertTypeToString(ParameterType::Direction::IN,
                    *i->GetType().GetMetaType());
              }
            }

            if (i->GetType().ToString() == "array") {
              return GetStringFromElementType(i->GetType());
            }

            return ConvertTypeToString(ParameterType::Direction::IN,
                i->GetType());
          },
          [&]()->std::string {
            if (i->GetType().ToString() == "array") {
              std::string str;
              str += i->GetID();
              str += ", ";
              str += "int " + i->GetID() + "_size";
              return str;
            }
            return i->GetID();
          },
          [&]()->std::string {
            if ((i->GetType().IsUserDefinedType() ||
                 i->GetType().ToString() == "string" ||
                 i->GetType().ToString() == "bundle") ||
                ((i->GetType().ToString() == "list" ||
                  i->GetType().ToString() == "array") &&
                  (i->GetType().GetMetaType()->IsUserDefinedType() ||
                   i->GetType().GetMetaType()->ToString() == "list" ||
                   i->GetType().GetMetaType()->ToString() == "array" ||
                   i->GetType().GetMetaType()->ToString() == "string" ||
                   i->GetType().GetMetaType()->ToString() == "bundle")))
              return "!h || !" + i->GetID();

            return "!h";
          },
          [&]()->std::string {
            std::string str;
            str += NLine(1);
            str += GetSetterString(i->GetID(), i->GetType());
            return str;
          }));
  }
}

void CBodyGeneratorBase::GenStructureGetter(std::ofstream& stream,
                                            const Structure& st) {
  for (auto& i : st.GetElements().GetElms()) {
    if (i->GetType().ToString() == "list")
      continue;

    stream << SmartIndent(GenTemplateString(CB_STRUCT_GETTER,
          [&]()->std::string {
            return GetStructIdWithNamespace(st);
          },
          [&]()->std::string {
            return i->GetID();
          },
          [&]()->std::string {
            return GetStructIdWithNamespace(st);
          },
          [&]()->std::string {
            if (i->GetType().IsUserDefinedType())
              return GetParcelParamTypeString(i->GetType());

            if (i->GetType().ToString() == "array")
              return GetStringFromElementType(i->GetType()) + "*";

            return ConvertTypeToString(ParameterType::Direction::OUT,
                i->GetType());
          },
          [&]()->std::string {
            if (i->GetType().ToString() == "array") {
              std::string str;
              str += i->GetID();
              str += ", ";
              str += "int *" + i->GetID() + "_size";
              return str;
            }
            return i->GetID();
          },
          [&]()->std::string {
            if (i->GetType().ToString() == "array") {
              std::string str;
              str += "!";
              str += i->GetID();
              str += "|| ";
              str += "!" + i->GetID() + "_size";
              return str;
            }
            return "!" + i->GetID();
          },
          [&]()->std::string {
            std::string str;
            str += NLine(1);
            str += GetGetterString(i->GetID(), i->GetType());
            return str;
          }));
  }
}

void CBodyGeneratorBase::GenStructureIterator(std::ofstream& stream,
                                              const Structure& st) {
  for (auto& i : st.GetElements().GetElms()) {
    if (i->GetType().ToString() != "list")
      continue;

    stream << SmartIndent(GenTemplateString(CB_STRUCT_ITERATOR,
          [&]()->std::string {
            return GetStructIdWithNamespace(st);
          },
          [&]()->std::string {
            return i->GetID();
          },
          [&]()->std::string {
            return GetStructIdWithNamespace(st);
          },
          [&]()->std::string {
            if (i->GetType().GetMetaType()->IsUserDefinedType() ||
                i->GetType().GetMetaType()->ToString() == "list" ||
                i->GetType().GetMetaType()->ToString() == "array")
              return GetParcelParamTypeString(*i->GetType().GetMetaType());

            return ConvertTypeToString(ParameterType::Direction::IN,
                *i->GetType().GetMetaType());
          },
          [&]()->std::string {
            return i->GetID();
          },
          [&]()->std::string {
            std::string str;
            str += NLine(1);
            str += GetIteratorString(i->GetID(), i->GetType());
            return str;
          }));
  }
}

void CBodyGeneratorBase::GenStructureCloner(std::ofstream& stream,
                                            const Structure& st) {
  stream << SmartIndent(GenTemplateString(
        ReplaceAll(CB_STRUCT_CLONER, "##", GetStructIdWithNamespace(st)),
        [&]()->std::string {
          std::string str;
          for (auto& i : st.GetElements().GetElms()) {
            str += NLine(1);
            str += GetClonerString(i->GetID(), i->GetType(), st);
          }
          return str;
        }));
}

std::string CBodyGeneratorBase::GetParcelTypeString(const BaseType& type,
                                                    bool meta_type) {
  if (type.IsUserDefinedType())
    return "";

  if (type.ToString() == "list" ||
      type.ToString() == "array") {
    if (meta_type)
      return "";
    return "array_count";
  }

  return parcel_type_map_[type.ToString()];
}

std::string CBodyGeneratorBase::GetParcelWriteFunctionString(
    const BaseType& type, bool meta_type) {
  std::string str = "rpc_port_parcel_write";
  std::string parcel_type = GetParcelTypeString(type, meta_type);
  if (parcel_type != "")
    str += "_" + parcel_type;

  return str;
}

std::string CBodyGeneratorBase::GetParcelWriteString(const std::string& id,
                                                     const BaseType& type) {
  std::string str;
  const char parcel[] = "$$(parcel, $$);";
  const char ternary_operation[] = "## ? ## : \"\"";

  str += GenTemplateString(parcel,
      [&]()->std::string {
        return GetParcelWriteFunctionString(type);
      },
      [&]()->std::string {
        if (type.IsUserDefinedType())
          return "&h->" + id + ", h->" + id;
        if (type.ToString() == "list")
          return "g_list_length(h->" + id + ")";
        if (type.ToString() == "array")
          return "h->" + id + "_size";
        if (type.ToString() == "string")
          return ReplaceAll(ternary_operation, "##", "h->" + id);
        return "h->" + id;
      });

  if (type.ToString() == "list") {
    str += GenTemplateString(CB_WRITE_LIST_BLOCK,
        [&]()->std::string {
          return "h->" + id;
        },
        [&]()->std::string {
          return GetParcelParamTypeString(*type.GetMetaType());
        },
        [&]()->std::string {
          return GenTemplateString(parcel,
              [&]()->std::string {
                return GetParcelWriteFunctionString(*type.GetMetaType(), true);
              },
              [&]()->std::string {
                if (type.GetMetaType()->IsUserDefinedType() ||
                    type.GetMetaType()->ToString() == "list" ||
                    type.GetMetaType()->ToString() == "array")
                  return "&value->parcelable, value";
                if (type.GetMetaType()->ToString() == "bundle")
                  return "value";
                if (type.GetMetaType()->ToString() == "string")
                  return ReplaceAll(ternary_operation, "##", "value");
                return "*value";
              });
        });
  } else if (type.ToString() == "array") {
    str += GenTemplateString(CB_WRITE_ARRAY_BLOCK,
        [&]()->std::string {
          return "h->" + id + "_size";
        },
        [&]()->std::string {
          return GenTemplateString(parcel,
              [&]()->std::string {
                return GetParcelWriteFunctionString(*type.GetMetaType(), true);
              },
              [&]()->std::string {
                if (type.GetMetaType()->IsUserDefinedType() ||
                    type.GetMetaType()->ToString() == "list" ||
                    type.GetMetaType()->ToString() == "array")
                  return "&h->" + id + "[i]->parcelable, h->" + id +"[i]";
                if (type.GetMetaType()->ToString() == "string")
                  return ReplaceAll(ternary_operation, "##", "h->" + id
                      + "[i]");
                return "h->" + id + "[i]";
              });
        });
  }

  return str;
}

std::string CBodyGeneratorBase::GetParcelReadFunctionString(
    const BaseType& type, bool meta_type) {
  std::string str = "rpc_port_parcel_read";
  std::string parcel_type = GetParcelTypeString(type, meta_type);
  if (parcel_type != "")
    str += "_" + parcel_type;

  return str;
}

std::string CBodyGeneratorBase::GetParcelReadString(const std::string& id,
                                                    const BaseType& type) {
  std::string str;
  const char parcel[] = "$$(parcel, $$);";

  if (type.ToString() == "list") {
    str += GenTemplateString(CB_READ_LIST_BLOCK,
        [&]()->std::string {
          return GenTemplateString(parcel,
              [&]()->std::string {
                return GetParcelReadFunctionString(type);
              },
              [&]()->std::string {
                return "&len";
              });
        },
        [&]()->std::string {
          return GetParcelParamTypeString(*type.GetMetaType());
        },
        [&]()->std::string {
          std::string s;

          if (type.GetMetaType()->IsUserDefinedType() ||
              type.GetMetaType()->ToString() == "list" ||
              type.GetMetaType()->ToString() == "array") {
            s += GetConstructorString(*type.GetMetaType(), "value");
            s += GenTemplateString(CB_IF_STATEMENT_WITH_BRACES,
                [&]()->std::string {
                  return "!value";
                },
                [&]()->std::string {
                  std::string ss;
                  ss += "_E(\"Failed to create handle\");" + NLine(1);
                  ss += "return;";
                  return ss;
                });
            s += NLine(1);
            s += GenTemplateString(parcel,
                [&]()->std::string {
                  return GetParcelReadFunctionString(*type.GetMetaType(), true);
                },
                [&]()->std::string {
                  return "&value->parcelable, value";
                });
          } else if (type.GetMetaType()->ToString() == "string" ||
                     type.GetMetaType()->ToString() == "bundle") {
            s += GenTemplateString(parcel,
                [&]()->std::string {
                  return GetParcelReadFunctionString(*type.GetMetaType());
                },
                [&]()->std::string {
                  return "&value";
                });
          } else {
            s += "value = calloc(1, sizeof(*value));" + NLine(1);
            s += GenTemplateString(CB_IF_STATEMENT_WITH_BRACES,
                [&]()->std::string {
                  return "!value";
                },
                [&]()->std::string {
                  std::string ss;
                  ss += "_E(\"Out of memory\");" + NLine(1);
                  ss += "return;";
                  return ss;
                });
            s += NLine(1);
            s += GenTemplateString(parcel,
                [&]()->std::string {
                  return GetParcelReadFunctionString(*type.GetMetaType());
                },
                [&]()->std::string {
                  return "value";
                });
          }

          return s;
        },
        [&]()->std::string {
          return "h->" + id;
        },
        [&]()->std::string {
          return "h->" + id;
        });
  } else if (type.ToString() == "array") {
    str += GenTemplateString(ReplaceAll(CB_READ_ARRAY_BLOCK, "##", id),
        [&]()->std::string {
          return GenTemplateString(parcel,
              [&]()->std::string {
                return GetParcelReadFunctionString(type);
              },
              [&]()->std::string {
                return "&h->" + id + "_size";
              });
        },
        [&]()->std::string {
          return GetReturnTypeString(*type.GetMetaType());
        },
        [&]()->std::string {
          return GetErrorValue(*type.GetMetaType());
        },
        [&]()->std::string {
          std::string s;
          if (type.GetMetaType()->IsUserDefinedType() ||
              type.GetMetaType()->ToString() == "list" ||
              type.GetMetaType()->ToString() == "array") {
            s += GetConstructorString(*type.GetMetaType(), "value");
            s += GenTemplateString(CB_IF_STATEMENT_WITH_BRACES,
                [&]()->std::string {
                  return "!value";
                },
                [&]()->std::string {
                  std::string ss;
                  ss += "_E(\"Failed to create handle\");" + NLine(1);
                  ss += "return;";
                  return ss;
                });
            s += NLine(1);
            s += GenTemplateString(parcel,
                [&]()->std::string {
                  return GetParcelReadFunctionString(*type.GetMetaType(), true);
                },
                [&]()->std::string {
                  return "&value->parcelable, value";
                });
          } else {
            s += GenTemplateString(parcel,
                [&]()->std::string {
                  return GetParcelReadFunctionString(*type.GetMetaType());
                },
                [&]()->std::string {
                  return "&value";
                });
          }
          s += NLine(1);
          s += GetSetterString("h->" + id + "[i]", "value");
          return s;
        });
  } else {
    str += GenTemplateString(parcel,
        [&]()->std::string {
          return GetParcelReadFunctionString(type);
        },
        [&]()->std::string {
          if (type.IsUserDefinedType())
            return "&h->" + id + "h->" + id;
          return "&h->" + id;
        });
  }

  return str;
}

std::string CBodyGeneratorBase::GetFinalizeString(const std::string& id,
                                                  const BaseType& type) {
  std::string str;

  if (!type.IsUserDefinedType() &&
       type.ToString() != "list" &&
       type.ToString() != "array" &&
       type.ToString() != "string" &&
       type.ToString() != "bundle")
    return str;

  if (type.ToString() == "list") {
    str += GenTemplateString(CB_FINALIZE_LIST_BLOCK,
        [&]()->std::string {
          return "h->" + id;
        },
        [&]()->std::string {
          return GetParcelParamTypeString(*type.GetMetaType());
        },
        [&]()->std::string {
          return GenTemplateString(CB_IF_STATEMENT,
              [&]()->std::string {
                return "value";
              },
              [&]()->std::string {
                return GetDestructorString(*type.GetMetaType(),
                    "value", true) + NLine(1);
              });
        },
        [&]()->std::string {
          return "h->" + id;
        });
  } else if (type.ToString() == "array") {
    if (!type.GetMetaType()->IsUserDefinedType() &&
         type.GetMetaType()->ToString() != "list" &&
         type.GetMetaType()->ToString() != "array" &&
         type.GetMetaType()->ToString() != "string" &&
         type.GetMetaType()->ToString() != "bundle") {
      return GenTemplateString(CB_IF_STATEMENT,
              [&]()->std::string {
                return "h->" + id;
              },
              [&]()->std::string {
                return "free(h->" + id + ");";
              });
    }
    str += GenTemplateString(CB_FINALIZE_ARRAY_BLOCK,
        [&]()->std::string {
          return "h->" + id + "_size";
        },
        [&]()->std::string {
          return GenTemplateString(CB_IF_STATEMENT,
              [&]()->std::string {
                return "h->" + id + "[i]";
              },
              [&]()->std::string {
                return GetDestructorString(*type.GetMetaType(),
                    "h->" + id + "[i]", true) + NLine(1);
              });
        },
        [&]()->std::string {
          return "h->" + id;
        });
  } else {
    str += GenTemplateString(CB_IF_STATEMENT,
        [&]()->std::string {
          return "h->" + id;
        },
        [&]()->std::string {
          return GetDestructorString(type, "h->" + id) + NLine(1);
        });
  }

  return str;
}

std::string CBodyGeneratorBase::GetSetterString(const std::string& id,
                                                const BaseType& type) {
  std::string str;
  const char setter[] = "$$ = $$;\n";

  if (type.IsUserDefinedType() ||
      type.ToString() == "string" ||
      type.ToString() == "bundle") {
    str += GenTemplateString(CB_IF_STATEMENT_WITH_BRACES,
        [&]()->std::string {
          return "h->" + id;
        },
        [&]()->std::string {
          std::string s;
          s += GetDestructorString(type, "h->" + id) + NLine(1);
          s += GetSetterString("h->" + id, "NULL");
          return s;
        });
    str += NLine(1);

    str += NLine(1);
    str += GetSetterString(type, "h->" + id, id);
    str += GenTemplateString(CB_IF_STATEMENT_WITH_BRACES,
        [&]()->std::string {
          return "!h->" + id;
        },
        [&]()->std::string {
          std::string s;
          s += "_E(\"Failed to duplicate data\");" + NLine(1);
          s += "return -1;";
          return s;
        });
    str += NLine(1);
  } else if (type.ToString() == "list") {
    if (type.GetMetaType()->IsUserDefinedType() ||
        type.GetMetaType()->ToString() == "string" ||
        type.GetMetaType()->ToString() == "bundle" ||
        type.GetMetaType()->ToString() == "list" ||
        type.GetMetaType()->ToString() == "array") {
      str += GenTemplateString(setter,
          [&]()->std::string {
            return "h->" + id;
          },
          [&]()->std::string {
            if (type.GetMetaType()->ToString() == "string")
              return "g_list_append(h->" + id + ", strdup(" + id + "))";
            return "g_list_append(h->" + id + ", " + id + ")";
          });
    } else {
      str += GenTemplateString(CB_SETTER_LIST_BLOCK,
          [&]()->std::string {
            return GetParcelParamTypeString(*type.GetMetaType());
          },
          [&]()->std::string {
            std::string s;
            s += GenTemplateString(CB_IF_STATEMENT_WITH_BRACES,
                [&]()->std::string {
                  return "!value";
                },
                [&]()->std::string {
                  std::string s;
                  s += "_E(\"Out of memory\");" + NLine(1);
                  s += "return -1;";
                  return s;
                });
            s += NLine(1);
            s += GetSetterString("*value", id);
            return s;
          },
          [&]()->std::string {
            return "h->" + id;
          },
          [&]()->std::string {
              return "h->" + id;
          });
    }
  } else if (type.ToString() == "array") {
    str += GetFinalizeString(id, type) + NLine(1);
    str += GetSetterString("h->" + id, "NULL");
    str += NLine(1);
    str += GenTemplateString(ReplaceAll(CB_SETTER_ARRAY_BLOCK, "##", id),
        [&]()->std::string {
          return GetSetterString(*type.GetMetaType(),
              "h->" + id + "[i]", id + "[i]");
        });
  } else {
    str += GetSetterString(type, "h->" + id, id);
  }

  return str;
}

std::string CBodyGeneratorBase::GetGetterString(const std::string& id,
                                                const BaseType& type) {
  std::string str;

  if (type.ToString() == "array") {
    str += GenTemplateString(ReplaceAll(CB_GETTER_ARRAY_BLOCK, "##", id),
        [&]()->std::string {
           return GetSetterString(*type.GetMetaType(),
              "(*" + id + ")[i]", "h->" + id + "[i]");
        });
  } else {
    if (type.IsUserDefinedType() ||
        type.ToString() == "string" ||
        type.ToString() == "bundle") {
      str += GenTemplateString(CB_IF_STATEMENT_WITH_BRACES,
          [&]()->std::string {
            return "!h->" + id;
          },
          [&]()->std::string {
            std::string s;
            s += "_E(\"Invalid parameter: h->" + id + " is NULL\");" + NLine(1);
            s += "return -1;";
            return s;
          });
      str += NLine(1);
      str += NLine(1);
    }

    if (type.IsUserDefinedType()) {
      str += GetSetterString(type, id, "h->" + id);
    } else {
      str += GetSetterString(type, "*" + id, "h->" + id);
    }

    if (type.IsUserDefinedType() ||
        type.ToString() == "string" ||
        type.ToString() == "bundle") {
      str += GenTemplateString(CB_IF_STATEMENT_WITH_BRACES,
          [&]()->std::string {
            return "*" + id + " == NULL";
          },
          [&]()->std::string {
            std::string s;
            s += "_E(\"Failed to duplicate " + id + "\");" + NLine(1);
            s += "return -1;";
            return s;
          });
      str += NLine(1);
    }
  }
  return str;
}

std::string CBodyGeneratorBase::GetIteratorString(const std::string& id,
                                                  const BaseType& type) {
  std::string str;

  str += GenTemplateString(CB_ITERATOR_BLOCK,
      [&]()->std::string {
        return "h->" + id;
      },
      [&]()->std::string {
        return GetParcelParamTypeString(*type.GetMetaType());
      },
      [&]()->std::string {
        if (type.GetMetaType()->IsUserDefinedType() ||
            type.GetMetaType()->ToString() == "list" ||
            type.GetMetaType()->ToString() == "array" ||
            type.GetMetaType()->ToString() == "string" ||
            type.GetMetaType()->ToString() == "bundle")
          return "value";

        return "*value";
      });

  return str;
}

std::string CBodyGeneratorBase::GetClonerString(const std::string& id,
                                                const BaseType& type,
                                                const Structure& st) {
  std::string str;

  if (type.IsUserDefinedType() ||
      type.ToString() == "string" ||
      type.ToString() == "bundle") {
    str += GenTemplateString(CB_IF_STATEMENT_WITH_BRACES,
        [&]()->std::string {
          return "h->" + id;
        },
        [&]()->std::string {
          std::string s;
          s += GetSetterString(type, "handle->" + id, "h->" + id);
          s += GenTemplateString(CB_IF_STATEMENT_WITH_BRACES,
              [&]()->std::string {
                return "!handle->" + id;
              },
              [&]()->std::string {
                std::string ss;
                ss += "_E(\"Failed to duplicate h->" + id + "\");" + NLine(1);
                ss += "rpc_port_" + GetStructIdWithNamespace(st)
                   + "_destroy(handle);" + NLine(1);
                ss += "return -1;";
                return ss;
              });
          return s;
        });
    str += NLine(1);
  } else  if (type.ToString() == "list") {
    str += GenTemplateString(CB_CLONER_LIST_BLOCK,
        [&]()->std::string {
          return "h->" + id;
        },
        [&]()->std::string {
          return GetParcelParamTypeString(*type.GetMetaType());
        },
        [&]()->std::string {
          return GetParcelParamTypeString(*type.GetMetaType());
        },
        [&]()->std::string {
          return GetStructIdWithNamespace(st);
        },
        [&]()->std::string {
          std::string s;
          if (type.GetMetaType()->IsUserDefinedType() ||
              type.GetMetaType()->ToString() == "list" ||
              type.GetMetaType()->ToString() == "array" ||
              type.GetMetaType()->ToString() == "string" ||
              type.GetMetaType()->ToString() == "bundle") {
            s += GetSetterString(*type.GetMetaType(),
                "new_value", "value");
            s += GenTemplateString(CB_IF_STATEMENT_WITH_BRACES,
                [&]()->std::string {
                  return "!new_value";
                },
              [&]()->std::string {
                std::string ss;
                ss += "_E(\"Failed to duplicate value\");" + NLine(1);
                ss += "rpc_port_" + GetStructIdWithNamespace(st)
                   + "_destroy(handle);" + NLine(1);
                ss += "return -1;";
                return ss;
              });
          } else {
            s += "new_value = calloc(1, sizeof(*new_value));" + NLine(1);
            s += GenTemplateString(CB_IF_STATEMENT_WITH_BRACES,
                [&]()->std::string {
                  return "!new_value";
                },
                [&]()->std::string {
                  std::string tmp;
                  tmp += "_E(\"Out of memory\");" + NLine(1);
                  tmp += "rpc_port_" + GetStructIdWithNamespace(st)
                      + "_destroy(handle);" + NLine(1);
                  tmp += "return -1;";
                  return tmp;
                });
            s += NLine(1);
            s += GetSetterString(*type.GetMetaType(),
                "*new_value", "*value");
          }
          s += NLine(1);
          return s;
        },
        [&]()->std::string {
          return "handle->" + id;
        },
        [&]()->std::string {
          return "handle->" + id;
        });
  } else if (type.ToString() == "array") {
    str += GenTemplateString(ReplaceAll(CB_CLONER_ARRAY_BLOCK, "##", id),
        [&]()->std::string {
          return GetStructIdWithNamespace(st);
        },
        [&]()->std::string {
           return GetSetterString(*type.GetMetaType(),
              "handle->" + id + "[i]", "h->" + id + "[i]");
        });
  } else {
    str += GetSetterString(type, "handle->" + id, "h->" + id);
  }

  return str;
}

std::string CBodyGeneratorBase::GetDestructorString(const BaseType& type,
                                                    const std::string& value,
                                                    bool container_value) {
  std::string str;

  if (type.IsUserDefinedType() ||
      type.ToString() == "list" ||
      type.ToString() == "array") {
    str += GetHandlePrefix() + GetFullNameFromType(type) +
        "_destroy(" + value + ");";
  } else if (type.ToString() == "bundle") {
    str += "bundle_free(" + value + ");";
  } else if (type.ToString() == "string" || container_value) {
    str += "free(" + value + ");";
  }

  return str;
}

std::string CBodyGeneratorBase::GetConstructorString(const BaseType& type,
                                                     const std::string& value) {
  std::string str;
  str += GetHandlePrefix() + GetFullNameFromType(type) +
      "_create(&" + value + ");" + NLine(1);
  return str;
}

std::string CBodyGeneratorBase::GetSetterString(const BaseType& type,
                                                const std::string& lvalue,
                                                const std::string& rvalue) {
  std::string str;

  if (type.IsUserDefinedType() ||
      type.ToString() == "list" ||
      type.ToString() == "array") {
    str += GetHandlePrefix() + GetFullNameFromType(type) +
        "_clone(" + rvalue + ", &" + lvalue +");" + NLine(1);
  } else if (type.ToString() == "string")  {
    str += lvalue + " = strdup(" + rvalue + ");" + NLine(1);
  } else if (type.ToString() == "bundle") {
    str += lvalue + " = bundle_dup(" + rvalue + ");" + NLine(1);
  } else {
    str += GetSetterString(lvalue, rvalue);
  }

  return str;
}

std::string CBodyGeneratorBase::GetSetterString(const std::string& lvalue,
                                                const std::string& rvalue) {
  std::string str;
  str += lvalue + " = " + rvalue + ";";
  return str;
}

void CBodyGeneratorBase::GenIncludeHeader(std::ofstream& stream) {
  std::string str;
  str += "#include \"";
  str += FileName.substr(0, FileName.length() - 2);
  str += ".h\"";
  stream << NLine(1);
  stream << str;
  stream << NLine(1);
}

void CBodyGeneratorBase::GenLogTag(std::ofstream& stream,
                                   const std::string& log_tag) {
  GenTemplate(CB_LOG_TAG, stream,
      [&]()->std::string {
        return log_tag;
      });
}

void CBodyGeneratorBase::GenInterfaceEnumerations(std::ofstream& stream,
                                                  const Interface& inf) {
  GenInterfaceMethodEnumeration(stream, inf);
  GenInterfaceDelegateEnumeration(stream, inf);
}

bool CBodyGeneratorBase::HasDelegate(const Interface& inf) const {
  for (auto& i : inf.GetDeclarations().GetDecls()) {
    if (i->GetMethodType() == Declaration::MethodType::DELEGATE) {
      return true;
    }
  }

  return false;
}

void CBodyGeneratorBase::GenInterfaceDelegateEnumeration(
    std::ofstream& stream, const Interface& inf) {
  int count = 1;

  if (!HasDelegate(inf))
    return;

  stream << SmartIndent(GenTemplateString(CB_DELEGATE_ENUM,
        [&]()->std::string {
          return GetInterfaceIdWithNamespace(inf);
        },
        [&]()->std::string {
          std::string str;
          for (auto& i : inf.GetDeclarations().GetDecls()) {
            if (i->GetMethodType() != Declaration::MethodType::DELEGATE)
              continue;

            str += GenTemplateString(CB_DELEGATE_ENUM_FORMAT,
                [&]()->std::string {
                  return GetInterfaceIdWithNamespace(inf);
                },
                [&]()->std::string {
                  return i->GetID();
                },
                [&]()->std::string {
                  return std::to_string(count++);
                });
          }
          return str;
        }));
}

void CBodyGeneratorBase::GenInterfaceMethodEnumeration(
    std::ofstream& stream, const Interface& inf) {
  stream << SmartIndent(GenTemplateString(CB_METHOD_ENUM,
        [&]()->std::string {
          return GetInterfaceIdWithNamespace(inf);
        },
        [&]()->std::string {
          return GetInterfaceIdWithNamespace(inf);
        },
        [&]()->std::string {
          return GetInterfaceIdWithNamespace(inf);
        },
        [&]()->std::string {
          std::string str;
          for (auto& i : inf.GetDeclarations().GetDecls()) {
            if (i->GetMethodType() == Declaration::MethodType::DELEGATE)
              continue;
            str += GenTemplateString(CB_METHOD_ENUM_FORMAT,
                [&]()->std::string {
                  return GetInterfaceIdWithNamespace(inf);
                },
                [&]()->std::string {
                  return i->GetID();
                });
          }
          return str;
        }));
}

void CBodyGeneratorBase::GenLogDefinition(std::ofstream& stream) {
  stream << SmartIndent(CB_LOG_DEF);
}

}  // namespace tidl
